#!/bin/bash
# 脚本版本号和更新说明
VERSION="1.9"
UPDATE_NOTES="1.1版本：使用rsync的输出并更新进度条，同时显示正在备份的文件。\n1.2版本：新增升级系统选项，允许用户选择全部更新、局部更新或选择具体更新的系统文件。\n1.3版本：检测系统功能更新，增加对文件系统、系统文件完整性和用户配置一致性的检测，并将日志位置设置为/var/log/system_check.log。\n1.4版本：重新设计系统回滚功能，提供单独的快照创建和系统回滚选项。新增升级包管理功能，支持上传、校验和安装系统升级包。\n1.5版本：添加用户管理功能，包括安装和配置 Cockpit，支持用户登录、注册和权限管理。\n1.6版本：升级功能支持差分升级，使升级包尽可能小。\n1.7版本：升级系统客户端增加安全机制，防止升级后台和升级包被篡改。\n1.8版本：升级功能确保用户应用/配置的兼容性。\n1.9版本：能查询终端的版本及硬件信息。"

# 显示脚本版本号和更新说明
show_update_notes() {
    dialog --msgbox "更新说明：\n$UPDATE_NOTES" 15 50
}

# 检查并安装 dialog 和其他必要工具
install_tools() {
    if ! command -v dialog &> /dev/null; then
        echo "dialog 工具未安装，正在安装..."
        sudo apt-get install -y dialog
    fi

    if ! command -v pv &> /dev/null; then
        echo "pv 工具未安装，正在安装..."
        sudo apt-get install -y pv
    fi

    if ! command -v debsums &> /dev/null; then
        echo "debsums 工具未安装，正在安装..."
        sudo apt-get install -y debsums
    fi

    if ! command -v aide &> /dev/null; then
        echo "AIDE 工具未安装，正在安装..."
        sudo apt-get install -y aide
    fi

    if ! command -v afick &> /dev/null; then
        echo "Afick 工具未安装，正在安装..."
        sudo apt-get install -y afick
    fi

    if ! command -v expect &> /dev/null; then
        echo "expect 工具未安装，正在安装..."
        sudo apt-get install -y expect
    fi
    
    if ! command -v snapper &> /dev/null; then
        echo "snapper 工具未安装，正在安装..."
        sudo apt-get install -y snapper
        sudo snapper -c root create-config /
    fi

    if ! command -v cockpit &> /dev/null; then
        echo "cockpit 工具未安装，正在安装..."
        sudo apt-get install -y cockpit
        sudo systemctl enable --now cockpit.socket
    fi

    if ! command -v lshw &> /dev/null; then
        echo "lshw 工具未安装，正在安装..."
        sudo apt-get install -y lshw
    fi
}

# 获取当前系统版本
get_current_version() {
    CURRENT_KERNEL=$(uname -r)
    CURRENT_UBUNTU=$(lsb_release -d | awk -F"\t" '{print $2}')
}

# 备份系统的函数
backup_system() {
    BACKUP_BASE=$(dialog --inputbox "请输入备份基础路径：" 10 50 3>&1 1>&2 2>&3 3>&-)
    TIMESTAMP=$(date +%Y%m%d%H%M%S)
    BACKUP_PATH="$BACKUP_BASE/backup_$TIMESTAMP"

    if [ ! -d "$BACKUP_PATH" ]; then
        mkdir -p "$BACKUP_PATH"
    fi

    dialog --msgbox "系统备份将包含以下内容：\n/boot\n/etc\n/home\n/root\n/var\n" 10 50

    total_size=$(du -sb /boot /etc /home /root /var 2>/dev/null | awk '{sum += $1} END {print sum}')

    # 创建一个临时文件用于记录 `rsync` 输出
    RSYNC_LOG=$(mktemp)

    # 启动进度条显示
    (
        while :; do
            sleep 1
            transferred=$(du -sb "$BACKUP_PATH" 2>/dev/null | awk '{print $1}')
            percent=$(awk "BEGIN {printf \"%.2f\", ($transferred/$total_size)*100}")
            echo "XXX"
            echo "$percent"
            echo "正在备份：$transferred bytes / $total_size bytes ($percent%)"
            echo "XXX"
        done
    ) | dialog --title "备份系统" --gauge "请等待..." 20 70 0 &

    PID=$!

    # 使用 `rsync` 和 `pv` 执行备份
    sudo rsync -aAXv /boot /etc /home /root /var "$BACKUP_PATH" >"$RSYNC_LOG" 2>&1

    # 停止进度条
    kill $PID

    # 显示 `rsync` 输出日志
    dialog --textbox "$RSYNC_LOG" 20 70

    # 删除临时文件
    rm -f "$RSYNC_LOG"

    dialog --msgbox "备份创建成功！存储在：$BACKUP_PATH" 5 40
}

# 全部更新的函数
full_update() {
    LOG_FILE=$(mktemp)
    dialog --title "系统更新中" --tailbox "$LOG_FILE" 20 70 &
    TAIL_PID=$!

    {
        sudo apt-get update
        sudo apt-get upgrade -y
        sudo apt-get dist-upgrade -y
        sudo apt-get autoremove -y
        sudo apt-get autoclean -y

        sudo apt-get install -y update-manager-core

        # 确保 release-upgrades 配置为 prompt=normal 或 lts
        sudo sed -i 's/^Prompt=.*/Prompt=normal/' /etc/update-manager/release-upgrades

        sudo do-release-upgrade -f DistUpgradeViewNonInteractive
    } > "$LOG_FILE" 2>&1

    kill $TAIL_PID

    if grep -q "An error occurred" "$LOG_FILE"; then
        dialog --msgbox "升级过程中出现错误，请查看日志：$LOG_FILE" 10 50
        return 1
    fi

    NEW_KERNEL=$(uname -r)
    NEW_UBUNTU=$(lsb_release -d | awk -F"\t" '{print $2}')
    NEW_UBUNTU_FULL=$(lsb_release -a 2>&1)
    dialog --msgbox "系统和内核更新成功！\n新系统版本：$NEW_UBUNTU\n新内核版本：$NEW_KERNEL\n\n详细信息：\n$NEW_UBUNTU_FULL" 15 60

    dialog --yesno "更新完成，需要重启系统以使更改生效。是否立即重启？" 10 50
    response=$?
    if [ $response -eq 0 ]; then
        sudo reboot
    else
        dialog --msgbox "请稍后手动重启系统以使更改生效。" 10 50
    fi
}

# 局部更新的函数
partial_update() {
    LOG_FILE=$(mktemp)
    dialog --title "局部更新中" --tailbox "$LOG_FILE" 20 70 &
    TAIL_PID=$!

    {
        sudo apt-get update
        sudo apt-get install -y linux-generic linux-headers-generic linux-image-generic initramfs-tools
        sudo update-initramfs -u
    } > "$LOG_FILE" 2>&1

    kill $TAIL_PID

    NEW_KERNEL=$(uname -r)
    NEW_UBUNTU=$(lsb_release -d | awk -F"\t" '{print $2}')
    NEW_UBUNTU_FULL=$(lsb_release -a 2>&1)
    dialog --msgbox "系统主要文件更新成功！\n新系统版本：$NEW_UBUNTU\n新内核版本：$NEW_KERNEL\n\n详细信息：\n$NEW_UBUNTU_FULL" 15 60

    dialog --yesno "更新完成，需要重启系统以使更改生效。是否立即重启？" 10 50
    response=$?
    if [ $response -eq 0 ]; then
        sudo reboot
    else
        dialog --msgbox "请稍后手动重启系统以使更改生效。" 10 50
    fi
}

# 选择要更新的系统文件的函数
selective_update() {
    choices=$(dialog --checklist "选择要更新的系统文件：" 15 50 5 \
        1 "更新内核" off \
        2 "更新 initramfs" off \
        3 "更新 rootfs" off \
        3>&1 1>&2 2>&3)

    LOG_FILE=$(mktemp)
    dialog --title "选择性更新中" --tailbox "$LOG_FILE" 20 70 &
    TAIL_PID=$!

    {
        if [[ $choices == *"1"* ]]; then
            sudo apt-get update
            sudo apt-get install -y linux-generic linux-headers-generic linux-image-generic
        fi

        if [[ $choices == *"2"* ]]; then
            sudo apt-get update
            sudo apt-get install -y initramfs-tools
            sudo update-initramfs -u
        fi

        if [[ $choices == *"3"* ]]; then
            sudo apt-get update
            sudo apt-get install -y --reinstall $(dpkg -l | grep '^ii' | awk '{print $2}')
        fi
    } > "$LOG_FILE" 2>&1

    kill $TAIL_PID

    NEW_KERNEL=$(uname -r)
    NEW_UBUNTU=$(lsb_release -d | awk -F"\t" '{print $2}')
    NEW_UBUNTU_FULL=$(lsb_release -a 2>&1)
    dialog --msgbox "选择的系统文件更新成功！\n新系统版本：$NEW_UBUNTU\n新内核版本：$NEW_KERNEL\n\n详细信息：\n$NEW_UBUNTU_FULL" 15 60

    dialog --yesno "更新完成，需要重启系统以使更改生效。是否立即重启？" 10 50
    response=$?
    if [ $response -eq 0 ]; then
        sudo reboot
    else
        dialog --msgbox "请稍后手动重启系统以使更改生效。" 10 50
    fi
}

# 升级系统的函数
upgrade_system() {
    CHOICE=$(dialog --clear --backtitle "系统升级" \
                    --title "选择升级类型" \
                    --menu "选择一个选项：" 15 50 3 \
                    1 "全部更新" \
                    2 "局部更新" \
                    3 "选择要更新的系统文件" \
                    3>&1 1>&2 2>&3)

    case $CHOICE in
        1)
            full_update
            ;;
        2)
            partial_update
            ;;
        3)
            selective_update
            ;;
    esac
}

# 重启系统的函数
reboot_system() {
    dialog --infobox "正在重启系统..." 5 40
    sudo reboot
}

# 检测系统的函数
check_system() {
    install_tools

    LOG_FILE="/var/log/system_check.log"
    dialog --title "检测系统中" --tailbox "$LOG_FILE" 20 70 &
    TAIL_PID=$!

    {
        echo "正在检查文件系统完整性..."
        sudo touch /forcefsck
        echo "文件系统检查将在下次启动时进行。" >> "$LOG_FILE"

        echo "正在检查系统文件完整性..."
        sudo debsums -s 2>/dev/null | tee -a "$LOG_FILE"

        echo "正在初始化 AIDE 数据库..."
        expect <<EOF | tee -a "$LOG_FILE"
spawn sudo aideinit
expect "Overwrite existing"
send "y\r"
expect eof
EOF

        echo "正在检查 AIDE 系统完整性..."
        sudo aide.wrapper --check 2>/dev/null | tee -a "$LOG_FILE"

        echo "正在检查用户配置一致性..."
        find /etc -type f -exec sha256sum {} + | sort > /tmp/before_upgrade_checksum

        echo "检测完成，正在验证用户配置一致性..."
        find /etc -type f -exec sha256sum {} + | sort > /tmp/after_upgrade_checksum
        if cmp -s /tmp/before_upgrade_checksum /tmp/after_upgrade_checksum; then
            echo "用户应用/配置一致性验证成功！"
        else
            echo "用户应用/配置一致性验证失败，请检查配置文件。"
        fi
        rm /tmp/before_upgrade_checksum /tmp/after_upgrade_checksum
    } > "$LOG_FILE" 2>&1

    kill $TAIL_PID
    dialog --msgbox "系统检测完成，请查看日志：$LOG_FILE" 10 50
}

# 创建快照的函数
create_snapshot() {
    install_tools
    sudo snapper -c root create --type pre --description "Before rollback"
    dialog --msgbox "创建快照成功。" 5 40
}

# 回滚系统的函数
rollback_system() {
    dialog --yesno "是否确认进行系统回滚？" 10 50
    response=$?
    if [ $response -eq 0 ]; then
        PRE_SNAPSHOT=$(sudo snapper -c root list | grep "pre" | tail -n 1 | awk '{print $1}')
        POST_SNAPSHOT=$(sudo snapper -c root list | grep "post" | tail -n 1 | awk '{print $1}')

        if [ -n "$PRE_SNAPSHOT" ] && [ -n "$POST_SNAPSHOT" ]; then
            sudo snapper -c root undochange ${PRE_SNAPSHOT}..${POST_SNAPSHOT}
            dialog --msgbox "系统回滚成功！" 10 50
        else
            dialog --msgbox "找不到有效的快照进行回滚。" 10 50
        fi
    else
        dialog --msgbox "取消回滚操作。" 10 50
    fi
}

# 校验升级包的函数
verify_package() {
    dialog --inputbox "请输入升级包的校验和 (SHA256)：" 10 50 2>checksum.txt
    EXPECTED_CHECKSUM=$(cat checksum.txt)
    ACTUAL_CHECKSUM=$(sha256sum "$UPGRADE_FILE" | awk '{print $1}')

    if [ "$EXPECTED_CHECKSUM" == "$ACTUAL_CHECKSUM" ]; then
        dialog --msgbox "校验成功！" 5 40
        return 0
    else
        dialog --msgbox "校验失败，请检查升级包。" 5 40
        return 1
    fi
}

# 升级包管理的函数
manage_upgrade_package() {
    dialog --fselect /home/ 10 50 2>file.txt
    UPGRADE_FILE=$(cat file.txt)

    verify_package "$UPGRADE_FILE"
    if [ $? -ne 0 ]; then
        return 1
    fi

    # 解压和安装升级包
    dialog --infobox "正在解压升级包..." 5 40
    tar -xf "$UPGRADE_FILE" -C /tmp
    dialog --infobox "正在安装升级包..." 5 40

    # 假设升级包包含一个 install.sh 脚本来执行安装
    if [ -f /tmp/install.sh ]; then
        sudo bash /tmp/install.sh
        if [ $? -ne 0 ]; then
            dialog --msgbox "升级失败，请检查升级脚本。" 5 40
            return 1
        fi
    else
        dialog --msgbox "升级包中未找到安装脚本，请检查升级包。" 5 40
        return 1
    fi

    dialog --msgbox "升级完成！" 5 40
    return 0
}

# 用户管理的函数
user_management() {
    dialog --msgbox "将通过Cockpit进行用户管理，浏览器将打开Cockpit界面。" 10 50
    xdg-open http://localhost:9090
}

# 查询终端信息的函数
terminal_info() {
    install_tools

    # 获取系统信息
    SYS_INFO=$(uname -a)
    
    # 获取发行版信息
    DISTRO_INFO=$(lsb_release -d | awk -F"\t" '{print $2}')
    
    # 获取CPU信息
    CPU_INFO=$(lscpu | grep "Model name" | awk -F: '{print $2}' | sed 's/^[ \t]*//')
    
    # 获取内存信息
    MEM_INFO=$(grep MemTotal /proc/meminfo | awk '{print $2/1024 " MB"}')
    
    # 获取硬盘信息
    DISK_INFO=$(lsblk -o NAME,SIZE -d | grep -E "^(sd|nvme|vd)" | awk '{print $1 ": " $2}')

    # 组合信息
    FORMATTED_INFO="系统信息：\n    $SYS_INFO\n\n发行版信息：\n    $DISTRO_INFO\n\nCPU信息：\n    $CPU_INFO\n\n内存信息：\n    $MEM_INFO\n\n硬盘信息：\n$(echo "$DISK_INFO" | sed 's/^/    /')"

    # 显示信息
    dialog --title "终端信息" --msgbox "$FORMATTED_INFO" 20 80
}

# 主菜单
while true; do
    CHOICE=$(dialog --clear --backtitle "Linux系统OTA升级 v$VERSION" \
                    --title "主菜单" \
                    --menu "选择一个选项：" 15 50 10 \
                    1 "备份系统" \
                    2 "升级系统" \
                    3 "升级包管理" \
                    4 "重启系统" \
                    5 "检测系统" \
                    6 "回滚系统" \
                    7 "用户管理" \
                    8 "终端信息" \
                    9 "更新说明" \
                    10 "退出" \
                    3>&1 1>&2 2>&3)

    case $CHOICE in
        1)
            backup_system
            ;;
        2)
            upgrade_system
            ;;
        3)
            manage_upgrade_package
            ;;
        4)
            reboot_system
            ;;
        5)
            check_system
            ;;
        6)
            CHOICE=$(dialog --clear --backtitle "系统回滚" \
                            --title "回滚系统" \
                            --menu "选择一个选项：" 15 50 2 \
                            1 "创建快照" \
                            2 "回滚系统" \
                            3>&1 1>&2 2>&3)

            case $CHOICE in
                1)
                    create_snapshot
                    ;;
                2)
                    rollback_system
                    ;;
            esac
            ;;
        7)
            user_management
            ;;
        8)
            terminal_info
            ;;
        9)
            show_update_notes
            ;;
        10)
            break
            ;;
    esac
done

clear
